/***************************************************************************
 *
 * COPYRIGHTHERE
 *
 * $Id: socket.c,v 1.1.4.1 2002/08/26 07:55:08 sasa Exp $
 *
 * Author  : Bazsi
 * Auditor :
 * Last audited version:
 * Notes:
 *
 ***************************************************************************/

#include <zorp/socket.h>
#include <zorp/cap.h>
#include <zorp/log.h>

/*+

  A thin interface around bind() using a ZSockAddr structure for
  socket address. It enables the NET_BIND_SERVICE capability (should be
  in the permitted set.

  Parameters:
    fd         fd to bind
    addr       address to bind to

  Returns:
    glib style I/O error

  +*/
GIOError
z_bind(gint fd, ZSockAddr *addr)
{
  cap_t saved_caps;
  GIOError rc;
  
  z_enter();
  saved_caps = cap_save();
  
  cap_enable(CAP_NET_BIND_SERVICE); /* for binding low numbered ports */
  cap_enable(CAP_NET_ADMIN); /* for binding non-local interfaces, transparent proxying */
  
  if (addr->sa_funcs && addr->sa_funcs->sa_bind_prepare)
    addr->sa_funcs->sa_bind_prepare(fd, addr);

  if (addr->sa_funcs && addr->sa_funcs->sa_bind)
    rc = addr->sa_funcs->sa_bind(fd, addr);
  else
    {
      if (addr && z_ll_bind(fd, &addr->sa, addr->salen) < 0)
        {
          z_log(NULL, CORE_ERROR, 1, "bind() failed; error='%m'");
          cap_restore(saved_caps);
          return G_IO_ERROR_UNKNOWN;
        }
      rc = G_IO_ERROR_NONE;
    }
  cap_restore(saved_caps);
  return rc;
}


/*+

  Accept a connection on the given fd, returning the newfd and the
  address of the client in a Zorp SockAddr structure.

  Parameters:
    fd         accept connection on this socket
    newfd      fd of the accepted connection
    addr       store the address of the client here

  Returns:
    glib style I/O error

  +*/
GIOError
z_accept(gint fd, gint *newfd, ZSockAddr **addr)
{
  char sabuf[1024];
  int salen = sizeof(sabuf);
  
  do
    {
      *newfd = z_ll_accept(fd, (struct sockaddr *) sabuf, &salen);
    }
  while (*newfd == -1 && errno == EINTR);
  if (*newfd != -1)
    {
      *addr = z_sockaddr_new((struct sockaddr *) sabuf, salen);
#if ENABLE_IPOPTIONS
      z_get_ip_options(newfd, *addr);
#endif
    }
  else if (errno == EAGAIN)
    {
      return G_IO_ERROR_AGAIN;
    }
  else
    {
      /*LOG
        This message indicates that accepting a connection was failed
        because of the given reason.
       */
      z_log(NULL, CORE_ERROR, 1, "accept() failed; fd=%d, error='%m'", fd);
      return G_IO_ERROR_UNKNOWN;
    }
  return G_IO_ERROR_NONE;
}

/*+

  Connect a socket using Zorp style ZSockAddr structure.

  Parameters:
    fd         socket to connect to
    remote     remote address

  Returns:
    glib style I/O error

  +*/
GIOError
z_connect(gint fd, ZSockAddr *remote)
{
  int rc;
  
  z_enter();

#if ENABLE_IPOPTIONS
  z_set_ip_options(fd, remote);
#endif
  do
    {
      rc = z_ll_connect(fd, &remote->sa, remote->salen);
    }
  while (rc == -1 && errno == EINTR);
  if (rc == -1)
    {
      if (errno != EINPROGRESS)
        {
          int saved_errno = errno;
      
          /*LOG
            This message indicates that establishing connection failed for
            the given reason.
           */
          z_log(NULL, CORE_ERROR, 1, "connect() failed; fd='%d', error='%m'", fd);
          errno = saved_errno;
        }
      z_leave();
      return G_IO_ERROR_UNKNOWN;
    }
  else
    {
      z_leave();
      return G_IO_ERROR_NONE;
    }
}

GIOError
z_listen(gint fd, gint backlog, gboolean accept_one)
{
  if (z_ll_listen(fd, backlog, accept_one) == -1)
    {
      return G_IO_ERROR_UNKNOWN;
    }
  return G_IO_ERROR_NONE;
}

GIOError
z_getsockname(gint fd, ZSockAddr **local_addr)
{
  char sabuf[1500];
  socklen_t salen = sizeof(sabuf);
  
  if (z_ll_getsockname(fd, (struct sockaddr *) sabuf, &salen) == -1)
    {
      return G_IO_ERROR_UNKNOWN;
    }
  *local_addr = z_sockaddr_new((struct sockaddr *) sabuf, salen);
  return G_IO_ERROR_NONE;
}

GIOError
z_getpeername(gint fd, ZSockAddr **peer_addr)
{
  char sabuf[1500];
  socklen_t salen = sizeof(sabuf);
  
  if (z_ll_getpeername(fd, (struct sockaddr *) sabuf, &salen) == -1)
    {
      return G_IO_ERROR_UNKNOWN;
    }
  *peer_addr = z_sockaddr_new((struct sockaddr *) sabuf, salen);
  return G_IO_ERROR_NONE;
}

/* low level functions */
gint
z_do_ll_bind(int fd, struct sockaddr *sa, socklen_t salen)
{
  return bind(fd, sa, salen);
}

gint
z_do_ll_accept(int fd, struct sockaddr *sa, socklen_t *salen)
{
  return accept(fd, sa, salen);
}

gint 
z_do_ll_connect(int fd, struct sockaddr *sa, socklen_t salen)
{
  return connect(fd, sa, salen);
}

gint 
z_do_ll_listen(int fd, gint backlog, gboolean accept_one G_GNUC_UNUSED)
{
  return listen(fd, backlog);
}

gint
z_do_ll_getsockname(int fd, struct sockaddr *sa, socklen_t *salen)
{
  return getsockname(fd, sa, salen);
}

gint 
z_do_ll_getpeername(int fd, struct sockaddr *sa, socklen_t *salen)
{
  return getpeername(fd, sa, salen);
}

ZSocketFuncs z_socket_funcs = 
{
  z_do_ll_bind,
  z_do_ll_accept,
  z_do_ll_connect,
  z_do_ll_listen,
  z_do_ll_getsockname,
  z_do_ll_getpeername
};

ZSocketFuncs *socket_funcs = &z_socket_funcs;

#if ENABLE_IPOPTIONS
static gboolean
z_get_ip_options(ZSocket *s, ZSockAddr *a)
{
  guchar buf[128]; /* ip options can be up to 40 bytes in size */
  guint buflen; 
  
  if (a->sa.sa_family != AF_INET)
    /* only AF_INET is supported */
    return FALSE;

  buflen = sizeof(buf);
  if (getsockopt(fd, SOL_IP, IP_OPTIONS, &buf, &buflen) == 0)
    {
      guchar newbuf[128];
      guchar *src = buf, *dst = newbuf;
      guint left = buflen, length;
      
      while (left)
        {
          if (*src == '\0')
            {
              /* EOOL */
              if (left > 1)
                {
                  /* hmm... this seems to be an invalid option list to me */
                  /*LOG
                    This message indicates that processing IP options failed,
                    because data was found after end-of-option-list. This might
                    be the result of a buggy TCP stack or an attack attempt.
                   */
                  z_log(NULL, CORE_ERROR, 1, "Invalid IP options, data after EOOL;");
                  return FALSE;
                }
              left--;
              break;
            }
          else if (*src == '\1')
            {
              /* NOP */
              left--;
            }
          else if (*src == 130 || *src == 133 || *src == 134)
            {
              /* RIPSO and CIPSO */
              
              /* this option is copied if present */
              if (left <= 1)
                {
                  /* no space for length field */
                  return FALSE;
                }
              length = *(src + 1);
              if (length > left)
                {
                  /* option too long */
                  return FALSE;
                }
              memcpy(dst, src, length);
              left -= length;
              dst += length;
              src += length;
            }
          else 
            {
              /* everything else is skipped */
              if (left <= 1)
                return FALSE;
              length = *(src + 1);
              left -= length;
              dst += length;
              src += length;
            }
        }
      z_sockaddr_inet_set_options(a, newbuf, dst - newbuf);
    }
  return TRUE;
}

static gboolean
z_set_ip_options(ZSocket *s, ZSockAddr *a)
{
  gpointer options = NULL;
  guint options_length;
    
  if (a->sa.sa_family != AF_INET)
    return FALSE;
  z_sockaddr_inet_get_options(a, &options, &options_length);
  if (options)
    {
      if (setsockopt(fd, SOL_IP, IP_OPTIONS, options, options_length) < 0)
        return FALSE;
    }
  return TRUE;    
}

#endif

