/***************************************************************************
 *
 * Copyright (c) 2000, 2001 BalaBit IT Ltd, Budapest, Hungary
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: zpython.c,v 1.32.2.6.2.4.2.3 2002/06/27 12:07:37 sasa Exp $
 *
 * Author  : Bazsi
 * Auditor : kisza
 * Last version: 1.22
 * Notes:
 *
 ***************************************************************************/

/* 
 * this module implements the interface with python 
 */
#include <zorp/zpython.h>
#include <zorp/log.h>
#include <zorp/io.h>
#include <zorp/policy.h>
#include <zorp/sysdep.h>

#include <zorp/pysockaddr.h>
#include <zorp/pylisten.h>
#include <zorp/pyreceive.h>
#include <zorp/pyconnect.h>
#include <zorp/pystream.h>
#include <zorp/pyproxy.h>
#include <zorp/pyipchains.h>
#include <zorp/pysockaddr.h>
#include <zorp/pyctrack.h>
#include <zorp/pytisauth.h>
#include <zorp/sockaddr.h>
#include <netdb.h>

PyThreadState *z_python_global_state;

/* exported python methods */

/*+

  Called by Python to send a message to the event log. Declaration
  in Python would be:

    def z_py_log(verbosity, priority, msg)

  Parameters:
    self       unused, required by Python calling mechanism
    args       arguments tuple

  Returns:
    PyNone

  +*/
static PyObject *
z_py_log(PyObject *self G_GNUC_UNUSED, PyObject *args)
{
  unsigned long verbosity;
  char *class, *msg;
  PyObject *py_session_id;
  gchar *session_id;

  if (!PyTuple_Check(args))
    {
      PyErr_SetString(PyExc_TypeError, "args must be a tuple");
      return NULL;
    }
  if (PyTuple_Size(args) == 3)
    {
      if (!PyArg_ParseTuple(args, "sis", &class, &verbosity, &msg))
        return NULL;
      session_id = NULL;
    }
  else
    {  
      if (!PyArg_ParseTuple(args, "Osis", &py_session_id, &class, &verbosity, &msg))
        return NULL;
      if (py_session_id == Py_None)
        {
          session_id = NULL;
        }
      else if (PyString_Check(py_session_id))
        {
          session_id = PyString_AsString(py_session_id);
        }
      else
        {
          PyErr_SetString(PyExc_TypeError, "Session ID must be string or None");
          return NULL;
        }


    }
    
  /*NOLOG*/
  z_log(session_id, class, verbosity, "%s", msg);
  Py_XINCREF(Py_None);
  return Py_None;
}

/*+
  +*/

static PyObject *
z_py_quit(PyObject *self G_GNUC_UNUSED, PyObject *args)
{
  int exit_code;
  
  if (!PyArg_ParseTuple(args, "i", &exit_code))
    return NULL;
  z_main_quit(exit_code);
  Py_XINCREF(Py_None);
  return Py_None;
}

static PyObject *
z_py_close_fd(PyObject *self G_GNUC_UNUSED, PyObject *args)
{
  int fd;
  
  if (!PyArg_ParseTuple(args, "i", &fd))
    return NULL;
  /*LOG
    This debug message indicates that the given fd is closed from Python.
   */
  z_log(NULL, CORE_DEBUG, 8, "closing fd from Python; fd='%d'", fd);
  close(fd);
  Py_XINCREF(Py_None);
  return Py_None;
}

/*+

  Query and return the original destination of a redirected connection.
  Python declaration would be:

    def z_py_get_transparent_dest(fd)

  Parameters:
    self       unused, required by Python calling mechanism
    args       arguments tuple

  Returns:
    ZorpSockAddr

  +*/
static PyObject *
z_py_get_transparent_dest(PyObject *self G_GNUC_UNUSED, PyObject *args)
{
  char buf[128];
  struct sockaddr *sa = (struct sockaddr *) buf;
  socklen_t salen = sizeof(buf);
  ZSockAddr *zsa;
  PyObject *res;
  int fd;
  
  if (!PyArg_ParseTuple(args, "i", &fd))
    return NULL;

  if (!z_get_transparent_dest(fd, sa, &salen))
    {
      PyErr_SetString(PyExc_IOError, "Error getting transparent destination");
      return NULL;
    }
  zsa = z_sockaddr_new(sa, salen);
  if (!zsa)
    {
      Py_XINCREF(Py_None);
      return Py_None;
    }
  res = z_py_zorp_sockaddr_new(zsa);
  z_sockaddr_unref(zsa);
  return res;
}

static PyObject *
z_py_get_sock_name(PyObject *self G_GNUC_UNUSED, PyObject *args)
{
  gchar sabuf[1500];
  gint sabuflen = sizeof(sabuf);
  gint fd;

  if (!PyArg_ParseTuple(args, "i", &fd))
    return NULL;
    
  if (getsockname(fd, (struct sockaddr *) sabuf, &sabuflen) < 0)
    {
      /*LOG
        This message indicates that the getsockname() system call failed.
       */
      z_log(NULL, CORE_ERROR, 4, "getsockname() failed; fd='%d', error='%m'", fd);
      Py_XINCREF(Py_None);
      return Py_None;
    }
  else
    {
      ZSockAddr *local = z_sockaddr_new((struct sockaddr *) sabuf, sabuflen);
      PyObject *res = z_py_zorp_sockaddr_new(local);
      
      z_sockaddr_unref(local);
      return res;
    }
}

static PyObject *
z_py_socket_pair(PyObject *self G_GNUC_UNUSED, PyObject *args)
{
  int domain, type, proto = 0;
  int result[2];
  
  if (!PyArg_ParseTuple(args, "ii|i", &domain, &type, &proto))
    {
      return NULL;
    }
  if (socketpair(domain, type, proto, result) == -1)
    {
      PyErr_SetString(PyExc_IOError, "I/O error during socketpair.");
      return NULL;
    }
  return z_policy_var_build("(ii)", result[0], result[1]);
}

static PyMethodDef zorp_funcs[] = 
{
  { "log", z_py_log, METH_VARARGS, NULL },
  { "getTransparentDestination", z_py_get_transparent_dest, METH_VARARGS, NULL },
  { "getSockName", z_py_get_sock_name, METH_VARARGS, NULL },
  { "closeFd", z_py_close_fd, METH_VARARGS, NULL },
  { "quit", z_py_quit, METH_VARARGS, NULL },
  { "socketPair", z_py_socket_pair, METH_VARARGS, NULL },
  { NULL, NULL, 0, NULL }
};

/*+

  Initialize the low level Python-Zorp interface. Called by the Python
  policy implementation.

  Parameters:
    none

  Returns:
    TRUE if initialization was successful

  +*/
gboolean
z_python_init(void)
{
  FILE *bootstrap;
  char buf[2048];

  if (getenv("PYTHONPATH") == NULL)
    {
      snprintf(buf, sizeof(buf), "PYTHONPATH=%s", ZORP_DATADIR "/pylib");
    }
  else
    {
      snprintf(buf, sizeof(buf), "PYTHONPATH=%s:%s", ZORP_DATADIR "/pylib", getenv("PYTHONPATH"));
    }
  putenv(buf);

  Py_Initialize();
  PyEval_InitThreads();

  
  if ((bootstrap = fopen(ZORP_POLICY_BOOT_FILE, "r")) == NULL)
    {
      /*LOG
        This message indicates that the policy.boot file couldn't be opened for
        reading.
       */
      z_log(NULL, CORE_ERROR, 0, "Error opening policy.boot file; file='%s'", ZORP_POLICY_BOOT_FILE);
      return FALSE;
    }
  PyRun_SimpleFile(bootstrap, ZORP_POLICY_BOOT_FILE);
  fclose(bootstrap);

  /* PyImport_AddModule("Zorp.Zorp"); */
  Py_InitModule("Zorp.Zorp", zorp_funcs);

  z_py_zorp_listen_init();
  z_py_zorp_receive_init();
  z_py_zorp_connect_init();
  z_py_zorp_stream_init();
  z_py_zorp_proxy_init();
  z_py_zorp_sockaddr_init();
  z_py_zorp_ipchains_init();
  z_py_zorp_ctracker_init();
  z_py_zorp_tisauth_init();
  
  z_python_global_state = PyEval_SaveThread();
  return TRUE;
}

/*+

  Called to deinitialize the Python-ZORP interface.

  Parameters:
    none
  
  Returns:
    TRUE if deinitialization was successful.

  +*/
gboolean
z_python_destroy(void)
{
  z_policy_lock(z_python_global_state);
  Py_Finalize();
  Py_Finalize();
  return TRUE;
}

/*+

  Lock the python interpreter, without setting the current thread pointer.

  Parameters:
    none

  Returns:
    none
  +*/
void 
z_python_lock(void)
{
  PyEval_AcquireLock();
}

/*+

  Unlock the python interpreter.

  Parameters:
    none

  Returns:
    none

  +*/
void 
z_python_unlock(void)
{
  PyEval_ReleaseLock();
}

